@@ -11,7 +11,7 @@ use Test2::Hub::Interceptor();
 use Test2::Hub::Interceptor::Terminator();
 
 use Carp qw/croak confess longmess/;
-use Scalar::Util qw/weaken blessed/;
+use Scalar::Util qw/blessed/;
 use Test2::Util qw/get_tid/;
 
 our @EXPORT_OK = qw{
@@ -38,6 +38,9 @@ our @EXPORT_OK = qw{
     test2_ipc_polling
     test2_ipc_disable_polling
     test2_ipc_enable_polling
+    test2_ipc_get_pending
+    test2_ipc_set_pending
+    test2_ipc_enable_shm
 
     test2_formatter
     test2_formatters
@@ -76,6 +79,9 @@ sub test2_ipc_drivers         { @{$INST->ipc_drivers} }
 sub test2_ipc_polling         { $INST->ipc_polling }
 sub test2_ipc_enable_polling  { $INST->enable_ipc_polling }
 sub test2_ipc_disable_polling { $INST->disable_ipc_polling }
+sub test2_ipc_get_pending     { $INST->get_ipc_pending }
+sub test2_ipc_set_pending     { $INST->set_ipc_pending(@_) }
+sub test2_ipc_enable_shm      { $INST->ipc_enable_shm }
 
 sub test2_formatter     { $INST->formatter }
 sub test2_formatters    { @{$INST->formatters} }
@@ -118,11 +124,26 @@ sub context {
         push @{$current->{_on_release}} => $params{on_release};
     }
 
-    return $current if $current && $current->{_depth} < $depth;
+    # Messy, but a quick way to return ASAP.
+    return $current if $current && $current->{_depth} < $depth && $current->{_canon_count} && $current->{_canon_count}++;
 
     # Handle error condition of bad level
-    _depth_error($current, [$pkg, $file, $line, $sub, $depth])
-        if $current;
+    if ($current) {
+        unless ($current->aborted) {
+            _canon_error($current, [$pkg, $file, $line, $sub, $depth])
+                unless $current->{_canon_count};
+
+            _depth_error($current, [$pkg, $file, $line, $sub, $depth])
+                unless $current->{_depth} < $depth;
+        }
+
+        if ($current->{_canon_count}) {
+            $current->{_canon_count} = 1;
+            $current->release;
+        }
+
+        delete $CONTEXTS->{$hid};
+    }
 
     # Directly bless the object here, calling new is a noticable performance
     # hit with how often this needs to be called.
@@ -139,17 +160,17 @@ sub context {
     # hit with how often this needs to be called.
     $current = bless(
         {
-            stack  => $stack,
-            hub    => $hub,
-            trace  => $trace,
-            _depth => $depth,
-            _err   => $@,
+            stack        => $stack,
+            hub          => $hub,
+            trace        => $trace,
+            _canon_count => 1,
+            _depth       => $depth,
             $params{on_release} ? (_on_release => [$params{on_release}]) : (),
         },
         'Test2::API::Context'
     );
 
-    weaken($CONTEXTS->{$hid} = $current);
+    $CONTEXTS->{$hid} = $current;
 
     $_->($current) for @$INIT_CBS;
 
@@ -163,8 +184,23 @@ sub context {
 }
 
 sub _depth_error {
-    my $ctx = shift;
-    my ($details) = @_;
+    _existing_error(@_, <<"    EOT");
+context() was called to retrieve an existing context, however the existing
+context was created in a stack frame at the same, or deeper level. This usually
+means that a tool failed to release the context when it was finished.
+    EOT
+}
+
+sub _canon_error {
+    _existing_error(@_, <<"    EOT");
+context() was called to retrieve an existing context, however the existing
+context has an invalid internal state (!_canon_count). This should not normally
+happen unless something is mucking about with internals...
+    EOT
+}
+
+sub _existing_error {
+    my ($ctx, $details, $msg) = @_;
     my ($pkg, $file, $line, $sub, $depth) = @$details;
 
     my $oldframe = $ctx->{trace}->frame;
@@ -173,10 +209,7 @@ sub _depth_error {
     my $mess = longmess();
 
     warn <<"    EOT";
-context() was called to retrieve an existing context, however the existing
-context was created in a stack frame at the same, or deeper level. This usually
-means that a tool failed to release the context when it was finished.
-
+$msg
 Old context details:
    File: $oldframe->[1]
    Line: $oldframe->[2]
@@ -193,10 +226,6 @@ Trace: $mess
 
 Removing the old context and creating a new one...
     EOT
-
-    my $hid = $ctx->{hub}->hid;
-    delete $CONTEXTS->{$hid};
-    $ctx->release;
 }
 
 sub release($;$) {
@@ -248,7 +277,7 @@ sub intercept(&) {
     $hub->finalize($trace, 1)
         if $ok
         && !$hub->no_ending
-        && !$hub->state->ended;
+        && !$hub->ended;
 
     return \@events;
 }
@@ -272,9 +301,6 @@ sub run_subtest {
     $hub->listen(sub { push @events => $_[1] });
     $hub->format(undef) if $buffered;
 
-    my $no_diag = defined($parent->get_todo) || $parent->parent_todo;
-    $hub->set_parent_todo($no_diag) if $no_diag;
-
     my ($ok, $err, $finished);
     T2_SUBTEST_WRAPPER: {
         # Do not use 'try' cause it localizes __DIE__, and does not preserve $@
@@ -309,9 +335,9 @@ sub run_subtest {
     $hub->finalize($trace, 1)
         if $ok
         && !$hub->no_ending
-        && !$hub->state->ended;
+        && !$hub->ended;
 
-    my $pass = $ok && $hub->state->is_passing;
+    my $pass = $ok && $hub->is_passing;
     my $e = $ctx->build_event(
         'Subtest',
         pass => $pass,
@@ -320,15 +346,17 @@ sub run_subtest {
         subevents => \@events,
     );
 
-    $e->set_diag([
-        $e->default_diag,
-        $ok ? () : ("Caught exception in subtest: $err"),
-    ]) unless $pass;
+    my $plan_ok = $hub->check_plan;
 
     $ctx->hub->send($e);
 
+    $ctx->diag("Caught exception in subtest: $err") unless $ok;
+
+    $ctx->diag("Bad subtest plan, expected " . $hub->plan . " but ran " . $hub->count)
+        if defined($plan_ok) && !$plan_ok;
+
     $ctx->release;
-    return $hub->state->is_passing;
+    return $pass;
 }
 
 1;
@@ -793,6 +821,28 @@ time a context is created.
 
 Turn off IPC polling.
 
+=item test2_ipc_enable_shm()
+
+Turn on IPC shm. Only some IPC drivers use this, and most will turn it on
+themselves.
+
+=item test2_ipc_set_pending($uniq_val)
+
+Tell other processes and events that an event is pending. C<$uniq_val> should
+be a unique value no other thread/process will generate.
+
+B<Note:> After calling this C<test2_ipc_get_pending()> will return 1. This is
+intentional, and not avoidable.
+
+=item $pending = test2_ipc_get_pending()
+
+This returns -1 if there is no way to check (assume yes)
+
+This returns 0 if there are (most likely) no pending events.
+
+This returns 1 if there are (likely) pending events. Upon return it will reset,
+nothing else will be able to see that there were pending events.
+
 =back
 
 =head2 MANAGING FORMATTERS
