@@ -3,35 +3,37 @@ use strict;
 use warnings;
 
 use Test2::Util::HashBase qw{
-    no_numbers no_header no_diag handles _encoding
+    no_numbers handles _encoding
 };
 
-sub OUT_STD()  { 0 }
-sub OUT_ERR()  { 1 }
-sub OUT_TODO() { 2 }
+sub OUT_STD() { 0 }
+sub OUT_ERR() { 1 }
 
 use Carp qw/croak/;
 
 use base 'Test2::Formatter';
 
 my %CONVERTERS = (
-    'Test2::Event::Ok'        => \&_ok_event,
-    'Test2::Event::Skip'      => \&_skip_event,
-    'Test2::Event::Note'      => \&_note_event,
-    'Test2::Event::Diag'      => \&_diag_event,
-    'Test2::Event::Bail'      => \&_bail_event,
-    'Test2::Event::Exception' => \&_exception_event,
-    'Test2::Event::Subtest'   => \&_subtest_event,
-    'Test2::Event::Plan'      => \&_plan_event,
+    'Test2::Event::Ok'        => 'event_ok',
+    'Test2::Event::Skip'      => 'event_skip',
+    'Test2::Event::Note'      => 'event_note',
+    'Test2::Event::Diag'      => 'event_diag',
+    'Test2::Event::Bail'      => 'event_bail',
+    'Test2::Event::Exception' => 'event_exception',
+    'Test2::Event::Subtest'   => 'event_subtest',
+    'Test2::Event::Plan'      => 'event_plan',
 );
 
+# Initial list of converters are safe for direct hash access cause we control them.
+my %SAFE_TO_ACCESS_HASH = %CONVERTERS;
+
 sub register_event {
     my $class = shift;
     my ($type, $convert) = @_;
     croak "Event type is a required argument" unless $type;
     croak "Event type '$type' already registered" if $CONVERTERS{$type};
-    croak "The second argument to register_event() must be a code reference"
-        unless $convert && ref($convert) eq 'CODE';
+    croak "The second argument to register_event() must be a code reference or method name"
+        unless $convert && (ref($convert) eq 'CODE' || $class->can($convert));
     $CONVERTERS{$type} = $convert;
 }
 
@@ -76,13 +78,17 @@ if ($^C) {
 sub write {
     my ($self, $e, $num) = @_;
 
-    my @tap = $self->event_tap($e, $num) or return;
+    my $type = ref($e);
+
+    my $converter = $CONVERTERS{$type} or return;
+    my @tap = $self->$converter($e, $self->{+NO_NUMBERS} ? undef : $num) or return;
 
     my $handles = $self->{+HANDLES};
-    my $nesting = $e->nested || 0;
+    my $nesting = ($SAFE_TO_ACCESS_HASH{$type} ? $e->{nested} : $e->nested) || 0;
     my $indent = '    ' x $nesting;
 
-    local($\, $", $,) = (undef, ' ', '');
+    # Local is expensive! Only do it if we really need to.
+    local($\, $,) = (undef, '') if $\ || $,;
     for my $set (@tap) {
         no warnings 'uninitialized';
         my ($hid, $msg) = @$set;
@@ -104,7 +110,7 @@ sub _open_handles {
     _autoflush($out);
     _autoflush($err);
 
-    return [$out, $err, $out];
+    return [$out, $err];
 }
 
 sub _autoflush {
@@ -125,43 +131,28 @@ sub event_tap {
     return $self->$converter($e, $num);
 }
 
-sub _ok_event {
+sub event_ok {
     my $self = shift;
     my ($e, $num) = @_;
 
     # We use direct hash access for performance. OK events are so common we
     # need this to be fast.
     my ($name, $todo) = @{$e}{qw/name todo/};
+    my $in_todo = defined($todo);
 
     my $out = "";
     $out .= "not " unless $e->{pass};
     $out .= "ok";
-    $out .= " $num" if defined $num;
+    $out .= " $num" if defined($num);
     $out .= " - $name" if defined $name;
+    $out .= " # TODO" if $in_todo;
+    $out .= " $todo" if length $todo;
 
-    if (defined $todo) {
-        $out .= " # TODO";
-        $out .= " $todo" if length $todo;
-    }
-
-    my @out = [OUT_STD, "$out\n"];
-
-    if ($e->{diag} && @{$e->{diag}}) {
-        my $diag_handle = ($todo || $e->diag_todo) ? OUT_TODO : OUT_ERR;
-
-        for my $diag (@{$e->{diag}}) {
-            chomp(my $msg = $diag);
-
-            $msg = "# $msg" unless $msg =~ m/^\n/;
-            $msg =~ s/\n/\n# /g;
-            push @out => [$diag_handle, "$msg\n"];
-        }
-    }
-
-    return @out;
+    # The primary line of TAP, if the test passed this is all we need.
+    return([OUT_STD, "$out\n"]);
 }
 
-sub _skip_event {
+sub event_skip {
     my $self = shift;
     my ($e, $num) = @_;
 
@@ -185,7 +176,7 @@ sub _skip_event {
     return([OUT_STD, "$out\n"]);
 }
 
-sub _note_event {
+sub event_note {
     my $self = shift;
     my ($e, $num) = @_;
 
@@ -196,22 +187,18 @@ sub _note_event {
     return [OUT_STD, "$msg\n"];
 }
 
-sub _diag_event {
+sub event_diag {
     my $self = shift;
     my ($e, $num) = @_;
-    return if $self->{+NO_DIAG};
 
     chomp(my $msg = $e->message);
     $msg =~ s/^/# /;
     $msg =~ s/\n/\n# /g;
 
-    return [
-        $e->todo ? OUT_TODO : OUT_ERR,
-        "$msg\n",
-    ];
+    return [OUT_ERR, "$msg\n"];
 }
 
-sub _bail_event {
+sub event_bail {
     my $self = shift;
     my ($e, $num) = @_;
 
@@ -223,49 +210,58 @@ sub _bail_event {
     ];
 }
 
-sub _exception_event {
+sub event_exception {
     my $self = shift;
     my ($e, $num) = @_;
     return [ OUT_ERR, $e->error ];
 }
 
-sub _subtest_event {
+sub event_subtest {
     my $self = shift;
     my ($e, $num) = @_;
 
-    my ($ok, @diag) = $self->_ok_event($e, $num);
+    # A 'subtest' is a subclass of 'ok'. Let the code that renders 'ok' render
+    # this event.
+    my ($ok, @diag) = $self->event_ok($e, $num);
 
-    return (
-        $ok,
-        @diag
-    ) unless $e->buffered;
+    # If the subtest is not buffered then the sub-events have already been
+    # rendered, we can go ahead and return.
+    return ($ok, @diag) unless $e->buffered;
 
+    # In a verbose harness we indent the diagnostics from the 'Ok' event since
+    # they will appear inside the subtest braces. This helps readability. In a
+    # non-verbose harness we do nto do this because it is less readable.
     if ($ENV{HARNESS_IS_VERBOSE}) {
-        $_->[1] =~ s/^/    /mg for @diag;
+        # index 0 is the filehandle, index 1 is the message we want to indent.
+        $_->[1] =~ s/^(.*\S)$/    $1/mg for @diag;
     }
 
+    # Add the trailing ' {' to the 'ok' line of TAP output.
     $ok->[1] =~ s/\n/ {\n/;
 
+    # Render the sub-events, we use our own counter for these.
     my $count = 0;
     my @subs = map {
-        $count++ if $_->isa('Test2::Event::Ok');
-        map { $_->[1] =~ s/^/    /mg; $_ } $self->event_tap($_, $count);
+        # Bump the count for any event that should bump it.
+        $count++ if $_->increments_count;
+
+        # This indents all output lines generated for the sub-events.
+        # index 0 is the filehandle, index 1 is the message we want to indent.
+        map { $_->[1] =~ s/^(.*\S)$/    $1/mg; $_ } $self->event_tap($_, $count);
     } @{$e->subevents};
 
     return (
-        $ok,
-        @diag,
-        @subs,
-        [OUT_STD(), "}\n"],
+        $ok,                # opening ok - name {
+        @diag,              #   diagnostics if the subtest failed
+        @subs,              #   All the inner-event lines
+        [OUT_STD(), "}\n"], # } (closing brace)
     );
 }
 
-sub _plan_event {
+sub event_plan {
     my $self = shift;
     my ($e, $num) = @_;
 
-    return if $self->{+NO_HEADER};
-
     my $directive = $e->directive;
     return if $directive && $directive eq 'NO PLAN';
 
@@ -311,21 +307,6 @@ This is what takes events and turns them into TAP.
 
     $tap->write($event, $number); # Output an event
 
-=head1 EXPORTS
-
-=over 4
-
-=item OUT_STD
-
-=item OUT_ERR
-
-=item OUT_TODO
-
-These are constants to identify filehandles. These constants are used by events
-to direct text to the correct filehandle.
-
-=back
-
 =head1 METHODS
 
 =over 4
@@ -336,24 +317,12 @@ to direct text to the correct filehandle.
 
 Use to turn numbers on and off.
 
-=item $bool = $tap->no_header($bool)
-
-=item $tap->set_no_header($bool)
-
-When true, the plan will not be rendered.
-
-=item $bool = $tap->no_diag
-
-=item $tap->set_no_diag($bool)
-
-When true, diagnostics will not be rendered.
-
 =item $arrayref = $tap->handles
 
 =item $tap->set_handles(\@handles);
 
 Can be used to get/set the filehandles. Indexes are identified by the
-C<OUT_STD, OUT_ERR, OUT_TODO> constants.
+C<OUT_STD> and C<OUT_ERR> constants.
 
 =item $encoding = $tap->encoding
 
@@ -375,7 +344,7 @@ you might want to write a custom event type that results in TAP output. In
 order to do this you use the C<register_event()> class method.
 
     package My::Event;
-    use Test2::Formatter::TAP qw/OUT_STD OUT_ERR/;
+    use Test2::Formatter::TAP;
 
     use base 'Test2::Event';
     use Test2::Util::HashBase accessors => [qw/pass name diag note/];
@@ -386,9 +355,9 @@ order to do this you use the C<register_event()> class method.
             my $self = shift;
             my ($e, $num) = @_;
             return (
-                [OUT_STD, "ok $num - " . $e->name . "\n"],
-                [OUT_ERR, "# " . $e->name . " " . $e->diag . "\n"],
-                [OUT_STD, "# " . $e->name . " " . $e->note . "\n"],
+                [Test2::Formatter::TAP::OUT_STD, "ok $num - " . $e->name . "\n"],
+                [Test2::Formatter::TAP::OUT_ERR, "# " . $e->name . " " . $e->diag . "\n"],
+                [Test2::Formatter::TAP::OUT_STD, "# " . $e->name . " " . $e->note . "\n"],
             );
         }
     );
@@ -397,6 +366,67 @@ order to do this you use the C<register_event()> class method.
 
 =back
 
+=head2 EVENT METHODS
+
+All these methods require the event itself. Optinally they can all except a
+test number.
+
+All methods return a list of array-refs. Each array-ref will have 2 items, the
+first is an integer identifying an output handle, the second is a string that
+should be written to the handle.
+
+=over 4
+
+=item @out = $TAP->event_ok($e)
+
+=item @out = $TAP->event_ok($e, $num)
+
+Process an L<Test2::Event::Ok> event.
+
+=item @out = $TAP->event_plan($e)
+
+=item @out = $TAP->event_plan($e, $num)
+
+Process an L<Test2::Event::Plan> event.
+
+=item @out = $TAP->event_note($e)
+
+=item @out = $TAP->event_note($e, $num)
+
+Process an L<Test2::Event::Note> event.
+
+=item @out = $TAP->event_diag($e)
+
+=item @out = $TAP->event_diag($e, $num)
+
+Process an L<Test2::Event::Diag> event.
+
+=item @out = $TAP->event_bail($e)
+
+=item @out = $TAP->event_bail($e, $num)
+
+Process an L<Test2::Event::Bail> event.
+
+=item @out = $TAP->event_exception($e)
+
+=item @out = $TAP->event_exception($e, $num)
+
+Process an L<Test2::Event::Exception> event.
+
+=item @out = $TAP->event_skip($e)
+
+=item @out = $TAP->event_skip($e, $num)
+
+Process an L<Test2::Event::Skip> event.
+
+=item @out = $TAP->event_subtest($e)
+
+=item @out = $TAP->event_subtest($e, $num)
+
+Process an L<Test2::Event::Subtest> event.
+
+=back
+
 =head1 SOURCE
 
 The source code repository for Test2 can be found at
